/*
 * Stella.c
 * 1/4/95
 * (c) 1993, 1995 by Stephen A. Ness (steve@mwc.com)
 * Cf. William S. Rubin, "Frank Stella", Museum of Modern Art, NY 1970.
 *
 * Display Frank Stella paintings with optical effects.
 * Usage: stella [ -dn ] [ -pn ] [ -w ]
 * Options:
 *	-dn	display for n seconds and then exit
 *	-pn	show picture n (1 to NPIX) only
 *	-w	leave white spaces white
 *
 * Type <SP> to pause/resume, 'q' or <Ctrl-C> to exit immediately,
 * 'a' to abort current effect.
 *
 * Uses libscreen and libvga:
 *	cc mwc.c -lscreen -lvga -lm
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "screen.h"
#include "vga.h"

#define	VERSION	"1.3"
#define	NPIX	9

/* Screen parameters. */
#define	COLORS	16
#define	BLACK	0
#define	WHITE	15

void	Delaware_Crossing();
void	Gran_Cairo();
void	Jaspers_Dilemma();
void	Kufa_Gate();
void	Marrakech();
void	Sacramento();
void	Sidi_Ifni();
void	Sinjerli_II();
void	Sinjerli_III();
void	recolor();
void	select();

/* Globals. */
int	npicture;			/* display picture n only */
int	sleepn;				/* sleep and exit */
#if	DEBUG
long	time0;				/* initial time */
#endif
int	wflag;				/* leave spaces white */

main(argc, argv) int argc; char *argv[];
{
	/* Process command line args. */
	if (argc > 1 && strncmp(argv[1], "-d", 2) == 0) {
		sleepn = atoi(&argv[1][2]);
		--argc;
		++argv;
	}
	if (argc > 1 && strncmp(argv[1], "-p", 2) == 0) {
		npicture = atoi(&argv[1][2]);
		--argc;
		++argv;
	}
	if (argc > 1 && strncmp(argv[1], "-w", 2) == 0) {
		++wflag;
		--argc;
		++argv;
	}
	if (argc > 1 || npicture < 0 || npicture > NPIX) {
		fprintf(stderr,
			"Stella: VGA V" VERSION " (c) Stephen A. Ness, 1993, 1995\n"
			"%s\n"
			"\n"
			"Usage: Stella [ -dn ] [ -pn ] [ -w ]\n"
			"Options:\n"
			"\t-dn\tdisplay for n seconds and then exit\n"
			"\t-pn\tshow picture n (1 to %d) only\n"
			"\t-w\tleave white spaces white (for purists)\n"
			"Type <space> to pause/resume, 'a' to abort current effect, 'q' to quit.\n"
			"\n"
			"Pictures by Frank Stella:\n"
			"\tDelaware Crossing (1961)\n"
			"\tGran Cairo (1962)\n"
			"\tJasper's Dilemma (1962-1963) [only left half shown]\n"
			"\tKufa Gate Shaped (1970) [Metropolitan Museum of Art poster]\n"
			"\tMarrakech (1964)\n"
			"\tSacramento Mall Proposal #4 (1978)\n"
			"\tSidi Ifni (1965)\n"
			"\tSinjerli II (1967)\n"
			"\tSinjerli III (1967)\n"
			"\n",
			scr_version, NPIX);
		exit(0);
	}
#if	DEBUG
	time0 = time(NULL);
	printf("0:\t");
#endif

	scr_init(G640x480x16);
	if (sleepn != 0) {
		/* Display the picture for sleepn seconds and exit. */
		select();
		scr_display(1);
		sleep(sleepn);
	} else {
		while (tty_check()) {
#if	DEBUG
			if (!memok()) {
				printf("memok failed!\n");
				exit(1);
			}
			printf("%ld:\t", time(NULL) - time0);
#endif
			select();
			recolor();
			scr_effect(RANDOM);
		}
	}
	scr_quit();
	exit(0);
}

/*
 * Delaware Crossing (1961).
 */
void
Delaware_Crossing()
{
	register int n, ulx, uly, lrx, lry, width, size;

	width = scr_xymin / 28;
	size = width * 28 - 1;			/* overall x/y width */
	ulx = (scr_xmax - size) / 2;
	uly = (scr_ymax - size) / 2;		/* offset */
	lrx = ulx + size;
	lry = uly + size;
	scr_clear(BLACK);
	scr_clip(ulx, uly, lrx, lry);
	scr_fillrectangle(ulx, uly, lrx, lry, 1);
	for (n = 0; n < 14 * width; n += width) {
		scr_wline(scr_x0 + n, scr_y0 + n, scr_x0 + n, lry,  2, WHITE);
		scr_wline(scr_x0 + n, scr_y0 + n, lrx, scr_y0 + n,  2, WHITE);
		scr_wline(scr_x0 - n, scr_y0 + n, scr_x0 - n, lry,  2, WHITE);
		scr_wline(scr_x0 - n, scr_y0 + n, ulx, scr_y0 + n,  2, WHITE);
		scr_wline(scr_x0 - n, scr_y0 - n, scr_x0 - n, uly,  2, WHITE);
		scr_wline(scr_x0 - n, scr_y0 - n, ulx, scr_y0 - n,  2, WHITE);
		scr_wline(scr_x0 + n, scr_y0 - n, scr_x0 + n, uly,  2, WHITE);
		scr_wline(scr_x0 + n, scr_y0 - n, lrx, scr_y0 - n,  2, WHITE);
	}
}

/*
 * Gran Cairo (1962).
 */
void
Gran_Cairo()
{
	static char colors[16] = { 1, 2, 3, 4, 5, 6, 5, 4, 3, 2, 1, 2, 3, 4, 5, 6 };
	register int x, y, n, size, width;

	width = scr_xymin / 31;
	size = width * 31;			/* overall x/y width */
	x = (scr_xmax - size) / 2;
	y = (scr_ymax - size) / 2;		/* offset */
	scr_clear(BLACK);
	scr_clip(x, y, x + size, y + size);
	for (n = 0; n < 16; n++) {
		scr_fillrectangle(x, y, x + size, y + size, WHITE);
		scr_fillrectangle(x + 2, y + 2, x + size - 2, y + size - 2, colors[n]);
		x += width;
		y += width;
		size -= 2 * width;
	}
}

/*
 * Jasper's Dilemma (1962-1963).
 * This is just the left half of the painting, the right half is in grays.
 */
void
Jaspers_Dilemma()
{
	register int x, y, whiteflag, n, c1, c2, c3, c4, oc2, size, width;
	char colors[60];

	width = scr_xymin / 28;
	size = width * 28;			/* overall x/y width */
	x = (scr_xmax - size) / 2;
	y = (scr_ymax - size) / 2;		/* offset */
	scr_clear(BLACK);
	scr_clip(x, y, x + size, y + size);

	/* Assign colors. */
	for (n = 0; n < 56; n++) {
again:
		c1 = colors[n] = rand() % (COLORS - 2) + 1;
		if ((n > 0 && c1 == colors[n-1])
		 || (n > 4 && c1 == colors[n-4]))
			goto again;		/* matched previous */
	}

	for (y = 0; y <= size / 2; y++) {
		if (y % width == 0) {
			/* Assign colors. */
			oc2 = c2;
			n = 4 * y / width;
			c1 = colors[n];
			c2 = colors[n + 1];
			c3 = colors[n + 2];
			c4 = colors[n + 3];
		}
		for (x = 0; x <= y; x++) {
			whiteflag = (y % width) == 0 || x == y || y == size / 2;
			scr0_color( x,  y) = (whiteflag) ? WHITE : c1;
			scr0_color(-x,  y) = (whiteflag) ? WHITE : c1;
			scr0_color( y,  x) = (whiteflag) ? WHITE : c2;
			scr0_color( y, -x) = (whiteflag) ? WHITE : c2;
			scr0_color(-x, -y) = (whiteflag) ? WHITE : c3;
			scr0_color( x, -y) = (whiteflag) ? WHITE : c3;
			scr0_color(-y,  x) = (whiteflag) ? WHITE : c4;
			scr0_color(-y, -x) = (whiteflag) ? WHITE : c4;
		}
		scr0_color(y - width, -y) = WHITE;
		for (x = y - width + 1; x <= y; x++) {
			if (x < 0 && x <= -y)
				continue;
			else if (y == size / 2 || x % width == 0)
				scr0_color(x, -y) = WHITE;
			else if (x < 0)
				continue;
			else if (x % width <= y % width)
				scr0_color(x, -y) = c2;
			else
				scr0_color(x, -y) = oc2;
		}
	}
}

/*
 * Kufa Gate Shaped (1970) [Metropolitan Museum of Art poster].
 */
void
Kufa_Gate()
{
	register int i, ulx, uly, lrx, lry, width, size;

	width = scr_xymin / 10;
	size = width * 10 - 2;			/* overall x/y width */
	ulx = (scr_xmax - size) / 2;
	uly = (scr_ymax - size) / 2;		/* offset */
	lrx = ulx + size - 1;
	lry = uly + size - 1;
	scr_clear(BLACK);
	scr_clip(ulx, uly, lrx, lry);
	scr_fillcircle(ulx, lry, size, 1);
	scr_fillcircle(lrx, lry, size, 1);
	scr_clip(ulx + width - 2, uly, lrx - width + 2, lry - width + 2);
	scr_fillcircle(ulx, lry, 9 * width, WHITE);
	scr_fillcircle(lrx, lry, 9 * width, WHITE);
	scr_clip(ulx + width, uly, lrx - width, lry - width);
	for (i = 9; i >= 2; i--) {
		if (i != 7 && i != 9) {
			scr_fillcircle(ulx, lry, i * width, WHITE);
			scr_fillcircle(ulx, lry, i * width-2, i);
		}
		scr_fillcircle(lrx, lry, i * width, WHITE);
		scr_fillcircle(lrx, lry, i * width-2, 5+i);
		if (i == 7 || i == 9) {
			scr_fillcircle(ulx, lry, i * width, WHITE);
			scr_fillcircle(ulx, lry, i * width-2, i);
		}
	}
	scr_clip(ulx, uly, lrx, lry);
}

/*
 * Marrakech (1964).
 */
void
Marrakech()
{
	register int x, y, flag, whiteflag, c1, c2, size, width;

	width = scr_xymin / 28;
	size = width * 28;			/* overall x/y width */
	x = (scr_xmax - size) / 2;
	y = (scr_ymax - size) / 2;		/* offset */
	scr_clear(BLACK);
	scr_clip(x, y, x + size, y + size);
	for (y = 0; y <= size / 2; y++)
	for (x = 0; x <= y; x++) {
		flag = (x / width ) % 2;
		whiteflag = (x % width) == 0 || x == y || y == size / 2;
		c1 =  (whiteflag) ? WHITE : (flag) ? 1 : 2;
		c2 =  (whiteflag) ? WHITE : (flag) ? 2 : 1;
		scr0_color( x,  y) = c1;
		scr0_color( y,  x) = c2;
		scr0_color(-x,  y) = c2;
		scr0_color(-y,  x) = c1;
		scr0_color(-x, -y) = c1;
		scr0_color(-y, -x) = c2;
		scr0_color( x, -y) = c2;
		scr0_color( y, -x) = c1;
	}
}

/*
 * Sacramento Mall Proposal #4 (1978).
 */
void
Sacramento()
{
	register int x, y, n, size, width;

	width = scr_xymin / 23;
	size = width * 23 - 1;			/* overall x/y width */
	width -= 4;				/* section width without border */
	x = (scr_xmax - size) / 2;
	y = (scr_ymax - size) / 2;		/* offset */
	scr_clear(BLACK);
	scr_clip(x, y, x + size, y + size);
	for (n = 0; n < 12; n++) {
		if (n != 0) {
			scr_fillrectangle(x, y, x + size, y + size, WHITE);
			x += 1;
			y += 1;
			size -= 2;
		}
		scr_fillrectangle(x, y, x + size, y + size, 1);
		x += 2;
		y += 2;
		size -= 4;
		scr_fillrectangle(x, y, x + size, y + size, WHITE);
		x += 1;
		y += 1;
		size -= 2;
		scr_fillrectangle(x, y, x + size, y + size, n + 2);
		x += width;
		y += width;
		size -= 2 * width;
	}
}

/*
 * Sidi Ifni (1965).
 */
void
Sidi_Ifni()
{
	register int x, y, n, maxx, width, whiteflag, flag;

	/* Draw the sections, computing only in the first octant. */
	width = scr_xymin / 22;			/* x width of each section */
	maxx = width * 11;			/* total x width from center */
	scr_clear(BLACK);
	scr_clip(scr_x0 - maxx, scr_y0 - maxx, scr_x0 + maxx, scr_y0 + maxx);
	for (n = 1; n < 22 * width; n++) {
		whiteflag = (n % width == 0);
		flag = (n / width) % 2;
		for (y = n / 2; y > 0; y--) {
			if ((x = n - y) >= maxx)
				break;
			scr0_color( x,  y) = (whiteflag) ? WHITE : (flag) ?  1 : 14;
			scr0_color( y,  x) = (whiteflag) ? WHITE : (flag) ? 14 :  4;
			scr0_color(-y,  x) = (whiteflag) ? WHITE : (flag) ?  5 : 14;
			scr0_color(-x,  y) = (whiteflag) ? WHITE : (flag) ? 14 :  7;
			scr0_color(-x, -y) = (whiteflag) ? WHITE : (flag) ? 11 : 14;
			scr0_color(-y, -x) = (whiteflag) ? WHITE : (flag) ? 14 : 12;
			scr0_color( y, -x) = (whiteflag) ? WHITE : (flag) ?  9 : 14;
			scr0_color( x, -y) = (whiteflag) ? WHITE : (flag) ? 14 :  8;
		}
	}

	/* Draw the white lines separating the sections. */
	for (n = 0; n <= maxx; n++) {
		scr0_color(-n,  maxx) = WHITE;
		scr0_color( n,  maxx) = WHITE;
		scr0_color(-n,     0) = WHITE;
		scr0_color( n,     0) = WHITE;
		scr0_color(-n, -maxx) = WHITE;
		scr0_color( n, -maxx) = WHITE;
		scr0_color( maxx, -n) = WHITE;
		scr0_color( maxx,  n) = WHITE;
		scr0_color(    0, -n ) = WHITE;
		scr0_color(    0,  n) = WHITE;
		scr0_color(-maxx, -n) = WHITE;
		scr0_color(-maxx,  n) = WHITE;
		scr0_color( n,  n) = WHITE;
		scr0_color( n, -n) = WHITE;
		scr0_color(-n, -n) = WHITE;
		scr0_color(-n,  n) = WHITE;
	}
}

/*
 * Sinjerli II (1967).
 */
void
Sinjerli_II()
{
	register int x, y, ysq, n, size, width, circle, color;
	char colors[80], c1, c2, c3, c4;
	int rsq[20];

	width = scr_xymin / 38;
	size = width * 38 + 2;
	x = (scr_xmax - size) / 2;
	y = (scr_ymax - size) / 2;		/* offset */
	scr_clear(BLACK);
	scr_clip(x, y, x + size, y + size);

	/* Assign colors. */
	for (n = 0; n < 40; n++) {
		colors[2 * n + 1] = WHITE;
again:
		color = colors[2 * n] = rand() % (COLORS - 2) + 1;
		if (n > 0 && color == colors[2 * n - 2])
			goto again;		/* matched previous */
		if (n % 10 == 9) {
			for (x = 1; x <= 9; x++)
				if (colors[2 * n - 2 * x] == color)
					goto again;	/* outer matched inner */
			for (x = n - 10; x > 0; x -= 10)
				if (colors[2 * x] == color)
					goto again;	/* outer matched outer */
		}
	}

	/* Compute squared radii of circles. */
	for (n = 0; n < 10; n++) {
		x = (2 * n + 1) * width - 2;
		rsq[2 * n] = x * x;
		x += 2;
		rsq[2 * n + 1] = x * x;
	}

	for (y = 0; y < size / 2; y++) {
		ysq = y * y;
		for (circle = x = 0; x <= y; x++) {
			while (x * x + ysq > rsq[circle])
				++circle;
			if (circle > 19)
				continue;
			if (x >= y-1) {
				c1 = c2 = c3 = c4 = WHITE;
			} else if (x >= y - 3 * width + 2 && circle < 19) {
				c1 = colors[18];
				c2 = colors[38];
				c3 = colors[58];
				c4 = colors[78];
			} else if (x >= y - 3 * width - 1 && circle < 17) {
				c1 = c2 = c3 = c4 = WHITE;
			} else {
				c1 = colors[ 0 + circle];
				c2 = colors[20 + circle];
				c3 = colors[40 + circle];
				c4 = colors[60 + circle];
			}
			scr0_color( x,  y) = c1;
			scr0_color(-x,  y) = c1;
			scr0_color( y,  x) = c2;
			scr0_color( y, -x) = c2;
			scr0_color(-x, -y) = c3;
			scr0_color( x, -y) = c3;
			scr0_color(-y, -x) = c4;
			scr0_color(-y,  x) = c4;
		}
	}
}

/*
 * Sinjerli III (1967).
 */
void
Sinjerli_III()
{
	register int x, y, x0, y0, n, r, rem, slice, size, width, color;
	char colors[25];
	double tang[7];

	/* This keeps same size as Sinjerli II. */
	width = scr_xymin / 38;
	size = width * 38 + 2;
	x = (scr_xmax - size) / 2;
	y = (scr_ymax - size) / 2;		/* offset */
	scr_clear(BLACK);
	scr_clip(x, y, x + size, y + size);
	x0 = scr_x0;
	y0 = scr_y0;

	/* Assign colors. */
	for (n = 0; n < 24; n++) {
again:
		color = colors[n] = rand() % (COLORS - 2) + 1;
		if (n == 0)
			colors[24] = colors[0];
		else if ((n == 12 && color == colors[0])
		      || color == colors[n-1]
		      || color == colors[(n % 12) * 12])
				goto again;
	}

	/* Compute tangents. */
	for (n = 1; n <= 6; n ++)
		tang[n] = tan(PI * (2. * n - 1.)/24.);

	scr_clip(x, y, x + size, y + size / 2);
	scr_fillcircle(x0, y0, size / 2, colors[0]);
	scr_clip(x, y + size / 2, x + size, y + size);
	scr_fillcircle(x0, y0, size / 2, colors[12]);
	scr_clip(x, y, x + size, y + size);
	scr_fillcircle(x0, y0, size / 2 - 2 * width, WHITE);

	r = size / 2 - 2 * width - 2;
	for (y = 0; y <= size / 2; y++) {
		rem = r * r - y * y;
		for (x = 0, slice = 6; x * x <= rem; x++) {
			if (slice > 0 && x * tang[slice] >= y)
				--slice;
			scr0_color( x,  y) = colors[     slice];
			scr0_color(-x,  y) = colors[12 - slice];
			scr0_color(-x, -y) = colors[12 + slice];
			scr0_color( x, -y) = colors[24 - slice];
		}
	}
	for (n = 0; n < 24; n++)
		scr_pline0(r, PI * (2 * n + 1) / 24, 2, WHITE);
	x = r + width;
	y = 2 * width;
	scr_fillrectangle(scr_x0 - x, scr_y0 - y, scr_x0 + x, scr_y0, colors[0]);
	scr_fillrectangle(scr_x0 - x, scr_y0, scr_x0 + x, scr_y0 + y, colors[12]);
	y = 2 * width + 1;
	scr_wline(scr_x0 - r, scr_y0 + y, scr_x0 + r, scr_y0 + y, 2, WHITE);
	scr_wline(scr_x0 - r, scr_y0 - y, scr_x0 + r, scr_y0 - y, 2, WHITE);
	scr_wline(scr_x0 - size, scr_y0, scr_x0 + size, scr_y0, 2, WHITE);
	x = (scr_xmax - size) / 2;
	y = (scr_ymax - size) / 2;
	scr_clip(x, y, x + size, y + size);
}

/*
 * Recolor the screen.
 * This version leaves BLACK as BLACK.
 * If wflag (for purists), it also leaves WHITE as WHITE.
 */
void
recolor()
{
	register int i, j;
	register char *s;
	char cbuf[COLORS];
	int tmp;

	/* Pick a randomized color set. */
	for (i = 0; i < COLORS; i++)
		cbuf[i] = i;
	for (i = COLORS, s = &cbuf[COLORS-1]; i > 2; i--, s--) {
		if (*s == WHITE && wflag)
			continue;
		/* Pick random j in [1, i-1], swap cbuf[i-1] <-> cbuf[j]. */
		/* This knows that BLACK==0, blech. */
		j = rand() % (i-1) + 1;		/* range [0, i - 1] */
		tmp = *s;
		*s = cbuf[j];
		cbuf[j] = tmp;
	}

	/* Remap the screen colors. */
	for (s = screen; s < screen + scr_npixels; s++)
		*s = cbuf[*s];
}

/*
 * Select a picture for display.
 * Kufa Gate Shaped and Sinjerli III take much longer to recompute than
 * the others, so they are saved in allocated buffers.
 * The others could be saved too, but on small systems (e.g. 4 MB)
 * this currently causes problems with COHERENT 4.2.r05.
 */
void
select()
{
	register int pic;
	static char *picture[NPIX];
	static int current = -1;
	static int ulx, uly, lrx, lry;

	static int count = 0;

	if (count++ == 0)
		pic = 2;
	else if (count == 1)
		pic = 7;
	else
		pic = rand() % NPIX;
	if (npicture != 0)
		pic = npicture - 1;		/* use specified picture */
	else if (current == -1)
		pic = rand() % NPIX;		/* random picture first time */
	else if (rand() % 2)
		pic = current;			/* keep current picture half the time */
	else					/* else random new picture */
		pic = (current + 1 + rand() % (NPIX-1)) % NPIX;

#if	DEBUG
	printf("select():\tpic=%d\n", pic);
#endif

	if (picture[pic] == NULL) {		/* new picture, compute it */

		switch(pic) {
		case 0:		Delaware_Crossing();	break;
		case 1:		Gran_Cairo();		break;
		case 2:		Jaspers_Dilemma();	break;
		case 3:		Kufa_Gate();		break;
		case 4:		Marrakech();		break;
		case 5:		Sacramento();		break;
		case 6:		Sidi_Ifni();		break;
		case 7:		Sinjerli_II();		break;
		case 8:		Sinjerli_III();		break;
		}
		if (current == -1) {		/* save clipping window dims */
			ulx = scr_ulx;
			uly = scr_uly;
			lrx = scr_lrx;
			lry = scr_lry;
		} else {
			/*
			 * Don't let the clipping window shrink,
			 * otherwise old pixels are left unchanged by effects.
			 */
			if (scr_ulx < ulx)
				ulx = scr_ulx;
			if (scr_uly < uly)
				uly = scr_uly;
			if (scr_lrx > lrx)
				lrx = scr_lrx;
			if (scr_lry > lry)
				lry = scr_lry;
			scr_clip(ulx, uly, lrx, lry);
		}
		/*
		 * The less time-consuming pictures are recomputed
		 * on the fly, cf. comment above.
		 */
		if ((pic == 3 || pic == 8)
		 && (picture[pic] = scr_alloc()) != NULL) {
			memcpy(picture[pic], screen, scr_npixels);
#if	DEBUG
			printf("\tpicture[%d]:\t0x%X\n", pic, picture[pic]);
#endif
		}
	} else if (current != pic)		/* restore computed picture */
		memcpy(screen, picture[pic], scr_npixels);
	current = pic;
}

/* end of Stella.c */
