/*
 * mwclogo.c
 * 2/4/94
 * (c) 1993, 1994 by Stephen A. Ness (steve@mwc.com)
 *
 * Display MWC logo with optical effects.
 * Usage: mwclogo [ -dn ] [ -tn ] [ foregr [ backgr ] ]
 * Options:
 *	-dn	display the logo for n seconds and then exit
 *	-tn	test scr_effect(n) only
 *	-vn	set global "value" to n, useful for debugging new effects
 *
 * Type <SP> to pause, <Ctrl-C> to exit immediately, 'a' to abort current
 * effect, or any other character to exit after current effect completes.
 *
 * Uses libscreen and libvga:
 *	cc mwc.c -lscreen -lvga -lm
 * 640x480x16 VGA screen is wired in for the moment.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "screen.h"
#include "vga.h"

#define	VERSION	"1.6"

#define	FOREGR	0
#define	BACKGR	15

void	mwclogo();
void	recolor();
void	usage();

/* Globals. */
int	backgr;				/* background color */
int	foregr;				/* foreground color */
int	sleepn;				/* sleep and exit */
EFFECT	test;				/* nonzero for testing effects */
int	value;				/* for testing effects */

main(argc, argv) int argc; char *argv[];
{
	/* Process command line args, basically useless. */
	while (argc > 1 && argv[1][0] == '-') {
		switch(argv[1][1]) {
		case 'd':
			sleepn = atoi(&argv[1][2]);
			break;
		case 't':
			test = atoi(&argv[1][2]);
			if (test < 0 || test > NEFFECTS) {
				fprintf(stderr, "mwclogo: -t arg must be in range [0, %d]\n",
					NEFFECTS);
				exit(1);
			}
			break;
		case 'v':
			value = atoi(&argv[1][2]);
			break;
		case 'V':
			fprintf(stderr, "mwclogo: V" VERSION "\n");
			fprintf(stderr, "%s\n", scr_version);
			exit(0);
		case '?':
		default:
			usage();
			break;
		}
		--argc;
		++argv;
	}
	foregr = (argc > 1) ? atoi(argv[1]) : FOREGR;
	backgr = (argc > 2) ? atoi(argv[2]) : BACKGR;
	if (argc > 3)
		usage();

	/* Display the picture.  The VGA mode is wired in for now. */
	scr_init(G640x480x16);
	mwclogo(foregr, backgr);		/* compute the bits */
	if (sleepn != 0) {
		scr_display(1);
		sleep(sleepn);
		scr_quit();
		exit(0);
	}
	while (tty_check()) {
		recolor();
		scr_effect((test) ? (EFFECT)test : RANDOM);
	}
	scr_quit();
	exit(0);
}

/*
 * Compute the MWC logo in screen with the given foreground and background.
 */
void
mwclogo(fore, back)
{
	/* Initialize screen to background color. */
	scr_clear(backgr);

	/* Circles. */
	scr_fillcircle(200, 135, 110, fore);	/* l top of M */
	scr_fillcircle(200, 135,  49, back);	/* white out excess */
	scr_fillcircle(200, 135,  30, fore);	/* l top of W */
	scr_fillcircle(360, 135, 110, fore);	/* r top of M */
	scr_fillcircle(360, 135,  49, back);	/* white out excess */
	scr_fillcircle(360, 135,  30, fore);	/* c top of W */
	scr_fillcircle(520, 135,  30, fore);	/* r top of W */

	scr_fillcircle(120, 345,  30, fore);	/* l bot of M */
	scr_fillcircle(280, 345, 110, fore);	/* l bot of W */
	scr_fillcircle(280, 345,  49, back);	/* white out excess */
	scr_fillcircle(280, 345,  30, fore);	/* c bot of M */
	scr_fillcircle(440, 345, 110, fore);	/* r bot of W */
	scr_fillcircle(440, 345,  49, back);	/* white out excess */
	scr_fillcircle(440, 345,  30, fore);	/* r bot of M */

	/* Rectangles. */
	scr_fillrectangle( 90, 135, 550, 345, back);	/* white out circle excess */
	scr_fillrectangle( 90, 135, 150, 345, fore);
	scr_fillrectangle(170, 135, 230, 345, fore);
	scr_fillrectangle(250, 135, 310, 345, fore);
	scr_fillrectangle(330, 135, 390, 345, fore);
	scr_fillrectangle(410, 135, 470, 345, fore);
	scr_fillrectangle(490, 135, 550, 345, fore);
}

/*
 * Recolor the current screen.
 * For the moment, this just remaps the foreground and background
 * to new randomly chosen values.
 * More generally, it should remap the entire current palette.
 */
void
recolor()
{
	register char *s;
	register int nf, nb, c1, c2;

	/* Choose c1 and c2 different from foregr, backgr, and each other. */
	do {
		c1 = rand() % scr_ncolors;
	} while (c1 == foregr || c1 == backgr);
	do {
		c2 = rand() % scr_ncolors;
	} while (c2 == foregr || c2 == backgr || c2 == c1);
	
	/*
	 * Visually, the most interesting transformations happen
	 * when only the foreground or only the background changes,
	 * or when foreground and background are reversed.
	 * These accordingly receive preference here over random coloration.
	 */
	switch(rand() % 6) {

	case 0:
		/* Reverse foreground and background. */
		nf = backgr;
		nb = foregr;
		break;

	case 1:
		/* Foreground becomes background, new foreground. */
		nb = foregr;
		nf = c1;
		break;

	case 2:
		/* Background becomes foreground, new background. */
		nf = backgr;
		nb = c1;
		break;

	case 3:		/* New foreground. */
		nb = backgr;
		nf = c1;
		break;

	case 4:		/* New background. */
		nf = foregr;
		nb = c1;
		break;

	case 5:
		nf = c1;
		nb = c2;
		break;
	}

	/* Remap the screen. */
	for (s = screen; s < screen + scr_npixels; s++) {
		if (*s == foregr)
			*s = nf;
		else if (*s == backgr)
			*s = nb;
	}
	foregr = nf;
	backgr = nb;
}

/*
 * Print usage message and exit.
 */
void
usage()
{
	fprintf(stderr,
		"Usage: mwclogo [ -dn ] [ -tn ] [ foregr [ backgr ] ]\n"
		"Options:\n"
		"\t-dn\tDisplay for n seconds, then exit\n"
		"\t-tn\tTest effect n only\n"
		"Type <SP> to pause/resume, 'a' to abort current effect, 'q' or <Esc> to quit.\n"
		);
	exit(1);
}

/* end of mwclogo.c */
