/*
 * scr_effect.c
 */

#include "screen.h"
#include <math.h>
#include <string.h>
#include <stdlib.h>

/*
 * Macros and manifest constants.
 */
#define	BLACK			0
#define	DIV(size, count)	(((size) + (count - 1)) / (count))
#define	NCHECK			20		/* keyboard input checks during dissolves */
#define	NSQPIX			10		/* pixels on square side in DISSSQ */
#define	RAND(l, u)		(rand() % ((u) + 1 - (l)) + (l))

/*
 * Magic numbers for dissolves.
 * Must be prime relative to scr_npixels if the dissolve is to hit every pixel.
 * The value of prime changes the optical effect considerably,
 * the values below were chosen by trial and error.
 * FIX_ME these values are for 640x480 screen, should be mode-dependent.
 */
int	primes[] = {
	1031,	1297,	2053,	2069,	4549,	10273,
	10289,	19919,	123457,	140111,	160967,	175069
};
#define	NDISS	(sizeof(primes) / sizeof(primes[0]))	/* number of dissolves	*/

/*
 * Redraw the screen using one of a plethora of optical effects.
 * This started out simple but jes' growed, adding new effects is fun.
 * This draws pixel by pixel, which is slow, but the resulting effect is nice.
 * The effect is performed only within the rectangular clipping window,
 * which is is a simple way to let the shape of the displayed object
 * determine the optical shape of the effect.
 * Out of range args are simply ignored.
 * The return value indicates whether the user interrupted,
 * which lets it skip subsequent recursive calls on compound effects.
 */
int
scr_effect(effect) EFFECT effect;
{
	register int x, y, n, count, i, dx, dy, nx, ny, prime, nsq, r, flag, iflag;
	int ulx, uly, lrx, lry, random;
	int *buf;
	div_t d;

	iflag = 1;			/* 0 if interrupted, 1 if not */
	if (effect == RANDOM)
		effect = (EFFECT)(1 + rand() % NEFFECTS);
#if	DEBUG
	printf("\tscr_effect():\teffect=%d\n", effect);
#endif

	switch(effect) {

	case DOWN:			/* wipe down */
		for (y = scr_uly; (iflag = tty_check()) && y <= scr_lry; y++)
		for (x = scr_ulx; x <= scr_lrx; x++)
			scr_pixel(x, y);
		break;

	case UP:			/* wipe up */
		for (y = scr_lry; (iflag = tty_check()) && y >= scr_uly; y--)
		for (x = scr_ulx; x <= scr_lrx; x++)
			scr_pixel(x, y);
		break;

	case UDTOC:			/* wipe up and down to center */
		count = (scr_ydim - 1) / 2;
		for (n = 0; (iflag = tty_check()) && n <= count; n++)
		for (x = scr_ulx; x <= scr_lrx; x++) {
			scr_pixel(x, scr_uly + n);
			scr_pixel(x, scr_lry - n);
		}
		break;

	case CTOUD:			/* wipe center to up and down */
		count = (scr_ydim - 1) / 2 + 1;
		for (n = 0; (iflag = tty_check()) && n <= count; n++)
		for (x = scr_ulx; x <= scr_lrx; x++) {
			scr_pixel(x, scr_y0 + n);
			scr_pixel(x, scr_y0 - n);
		}
		break;

	case LTOR:			/* wipe left to right */
		for (x = scr_ulx; (iflag = tty_check()) && x <= scr_lrx; x++)
		for (y = scr_uly; y <= scr_lry; y++)
			scr_pixel(x, y);
		break;

	case RTOL:			/* wipe right to left */
		for (x = scr_lrx; (iflag = tty_check()) && x >= scr_ulx; x--)
		for (y = scr_uly; y <= scr_lry; y++)
			scr_pixel(x, y);
		break;

	case LRTOC:			/* wipe left and right to center */
		count = (scr_xdim - 1) / 2;
		for (n = 0; (iflag = tty_check()) && n <= count; n++)
		for (y = scr_uly; y <= scr_lry; y++) {
			scr_pixel(scr_ulx + n, y);
			scr_pixel(scr_lrx - n, y);
		}
		break;

	case CTOLR:			/* wipe center to left and right */
		count = (scr_xdim - 1) / 2 + 1;
		for (n = 0; (iflag = tty_check()) && n <= count; n++)
		for (y = scr_uly; y <= scr_lry; y++) {
			scr_pixel(scr_x0 + n, y);
			scr_pixel(scr_x0 - n, y);
		}
		break;

	case UDN:			/* n-window wipe up to down */
		count = RAND(2, 10);		/* range [2, 10] windows */
		dy = DIV(scr_ydim, count);
		for (y = scr_uly; (iflag = tty_check()) && y < scr_uly + dy; y++)
		for (n = 0; n < dy * count; n += dy)
		for (x = scr_ulx; x <= scr_lrx; x++)
			scr_pixel(x, y + n);
		break;

	case LRN:			/* n-window wipe left to right */
		count = RAND(2, 10);		/* range [2, 10] windows */
		dx = DIV(scr_xdim, count);
		for (x = scr_ulx; (iflag = tty_check()) && x < scr_ulx + dx; x++)
		for (n = 0; n < dx * count; n += dx)
		for (y = scr_uly; y <= scr_lry; y++)
			scr_pixel(x + n, y);
		break;

	case UDSQN:			/* n x n-window wipe up to down */
		count = RAND(4, 12);		/* range [4, 12] windows */
		dx = DIV(scr_xdim, count);
		dy = DIV(scr_ydim, count);
		for (i = 0; iflag && (iflag = tty_check()) && i <= 1; i++)
		for (y = scr_uly; (iflag = tty_check()) && y < scr_uly + dy; y++)
		for (n = ny = 0; n < count; n++, ny += dy)
		for (x = scr_ulx; x < scr_ulx + dx; x++)
		for (nx = dx * ((n+i) % 2); nx < dx * count; nx += 2 * dx)
			scr_pixel(x + nx, y + ny);
		break;

	case LRSQN:			/* n x n-window wipe left to right */
		count = RAND(4, 12);		/* range [4, 12] windows */
		dx = DIV(scr_xdim, count);
		dy = DIV(scr_ydim, count);
		for (i = 0; iflag && (iflag = tty_check()) && i <= 1; i++)
		for (x = scr_ulx; (iflag = tty_check()) && x < scr_ulx + dx; x++)
		for (n = nx = 0; n < count; n++, nx += dx)
		for (y = scr_uly; y < scr_uly + dy; y++)
		for (ny = dy * ((n+i) % 2); ny < dy * count; ny += 2 * dy)
			scr_pixel(x + nx, y + ny);
		break;

	case DIAGUL:			/* wipe diagonally from upper left */
		for (n = 0; (iflag = tty_check()) && n <= scr_xdim + scr_ydim; n++)
		for (y = 0; y <= n; y++)
			scr_pixel(scr_ulx + n - y, scr_uly + y);
		break;

	case DIAGUR:			/* wipe diagonally from upper right */
		for (n = 0; (iflag = tty_check()) && n <= scr_xdim + scr_ydim; n++)
		for (y = 0; y <= n; y++)
			scr_pixel(scr_lrx - n + y, scr_uly + y);
		break;

	case DIAGLL:			/* wipe diagonally from lower left */
		for (n = 0; (iflag = tty_check()) && n <= scr_xdim + scr_ydim; n++)
		for (y = 0; y <= n; y++)
			scr_pixel(scr_ulx + n - y, scr_lry - y);
		break;

	case DIAGLR:			/* wipe diagonally from lower right */
		for (n = 0; (iflag = tty_check()) && n <= scr_xdim + scr_ydim; n++)
		for (y = 0; y <= n; y++)
			scr_pixel(scr_lrx - n + y, scr_lry - y);
		break;

	case SQIN:			/* square wipe from outside in */
		for (n = scr_xymax / 2; (iflag = tty_check()) && n >= 0; n--)
		for (y = -n; y <= n; y++) {
			scr_pixel(scr_x0 + y, scr_y0 - n);
			scr_pixel(scr_x0 - y, scr_y0 + n);
			scr_pixel(scr_x0 + n, scr_y0 + y);
			scr_pixel(scr_x0 - n, scr_y0 - y);
		}
		break;

	case SQOUT:			/* square wipe from inside out */
		for (n = 0; (iflag = tty_check()) && n <= scr_xymax / 2; n++)
		for (y = -n; y <= n; y++) {
			scr_pixel(scr_x0 + y, scr_y0 - n);
			scr_pixel(scr_x0 - y, scr_y0 + n);
			scr_pixel(scr_x0 + n, scr_y0 + y);
			scr_pixel(scr_x0 - n, scr_y0 - y);
		}
		break;

	case DIAGIN:			/* diagonal square wipe in */
		for (n = 0; (iflag = tty_check()) && n <= (scr_xdim + scr_ydim) / 2; n++)
		for (y = 0; y <= n && y < scr_ydim; y++) {
			scr_pixel(scr_ulx + n - y, scr_uly + y);
			scr_pixel(scr_lrx - n + y, scr_uly + y);
			scr_pixel(scr_ulx + n - y, scr_lry - y);
			scr_pixel(scr_lrx - n + y, scr_lry - y);
		}
		break;

	case DIAGOUT:			/* diagonal square wipe out */
		for (n = 0; (iflag = tty_check()) && n <= (scr_xdim + scr_ydim) / 2; n++)
		for (y = 0; y <= n; y++) {
			scr_pixel(scr_x0 - n + y, scr_y0 + y);
			scr_pixel(scr_x0 + n - y, scr_y0 - y);
			scr_pixel(scr_x0 - n + y, scr_y0 - y);
			scr_pixel(scr_x0 + n - y, scr_y0 + y);
		}
		break;

	case CIN:			/* circle in */
		/*
		 * For efficiency, keep an array of which gives
		 * the first x value to test for each given y value.
		 * Compute in the first quadrant, the others are the same.
		 */
		buf = alloca(scr_ydim * sizeof(int));	/* allocate temporary storage */
		r = (int)sqrt((double)((scr_xdim/2+1)*(scr_xdim/2+1)
				     + (scr_ydim/2+1)*(scr_ydim/2+1)));
		for (y = 0; y < scr_ydim / 2 + 1; y++)
			buf[y] = scr_xdim / 2 + 1;
		for (n = r; (iflag = tty_check()) && n >= 0; n--) {
			nsq = n * n;
			for (y = 0; y < scr_ydim / 2 + 1; y++)
			for (x = buf[y]; x >= 0; x--) {
				if (x*x + y*y < nsq)
					break;
				scr_pixel(scr_x0 + x, scr_y0 + y);
				scr_pixel(scr_x0 + x, scr_y0 - y);
				scr_pixel(scr_x0 - x, scr_y0 + y);
				scr_pixel(scr_x0 - x, scr_y0 - y);
				buf[y] = x;
			}
		}
		break;

	case COUT:			/* circle out */
		/* See comments on CIN above. */
		buf = alloca(scr_ydim * sizeof(int));	/* allocate temporary storage */
		r = (int)sqrt((double)((scr_xdim/2+1)*(scr_xdim/2+1)
				     + (scr_ydim/2+1)*(scr_ydim/2+1)));
		for (y = 0; y < scr_ydim / 2 + 1; y++)
			buf[y] = 0;
		for (n = 0; (iflag = tty_check()) && n <= r; n++) {
			nsq = n * n;
			for (y = 0; y < scr_ydim / 2 + 1; y++)
			for (x = buf[y]; x <= scr_xdim / 2 + 1; x++) {
				if (x*x + y*y > nsq)
					break;
				scr_pixel(scr_x0 + x, scr_y0 + y);
				scr_pixel(scr_x0 + x, scr_y0 - y);
				scr_pixel(scr_x0 - x, scr_y0 + y);
				scr_pixel(scr_x0 - x, scr_y0 - y);
				buf[y] = x;
			}
		}
		break;

	case CSQIN:			/* n x n circles in */
		/* See comments on CIN above. */
		buf = alloca(scr_ydim * sizeof(int));	/* allocate temporary storage */
		count = RAND(2, 10);		/* range [2, 10] windows */
		dx = DIV(scr_xdim, count);
		dy = DIV(scr_ydim, count);
		r = (int)sqrt((double)(dx*dx/4 + dy*dy/4)) + 1;
		for (y = 0; y < r + 1; y++)
			buf[y] = dx / 2;
		for (n = r; (iflag = tty_check()) && n >= 0; n--) {
			nsq = n * n;
			for (y = 0; y <= dy / 2; y++) {
				i = nsq - y * y;
				for (x = buf[y]; x >= 0 && x * x >= i; x--) {
					for (ny = scr_uly + dy / 2; ny < scr_uly + dy * count; ny += dy)
					for (nx = scr_ulx + dx / 2; nx < scr_ulx + dx * count; nx += dx) {
						scr_pixel(nx + x, ny + y);
						scr_pixel(nx + x, ny - y);
						scr_pixel(nx - x, ny + y);
						scr_pixel(nx - x, ny - y);
					}
					buf[y] = x;
				}
			}
		}
		break;

	case CSQOUT:			/* n x n circles out */
		/* See comments on CIN above. */
		buf = alloca(scr_ydim * sizeof(int));	/* allocate temporary storage */
		count = RAND(2, 10);		/* range [2, 10] windows */
		dx = DIV(scr_xdim, count);
		dy = DIV(scr_ydim, count);
		r = (int)sqrt((double)(dx*dx/4 + dy*dy/4)) + 1;
		for (y = 0; y < dy / 2 + 1; y++)
			buf[y] = 0;
		for (n = 0; (iflag = tty_check()) && n <= r; n++) {
			nsq = n * n;
			for (y = 0; y <= dy / 2 + 1; y++) {
				i = nsq - y * y;
				for (x = buf[y]; x <= dx / 2 + 1 && x * x <= i; x++) {
					for (ny = scr_uly + dy / 2; ny < scr_uly + dy * count; ny += dy)
					for (nx = scr_ulx + dx / 2; nx < scr_ulx + dx * count; nx += dx) {
						scr_pixel(nx + x, ny + y);
						scr_pixel(nx + x, ny - y);
						scr_pixel(nx - x, ny + y);
						scr_pixel(nx - x, ny - y);
					}
					buf[y] = x;
				}
			}
		}
		break;

	case DSQN:			/* n x n-window diagonal wipe */
		count = RAND(4, 12);		/* range [4, 12] windows */
		dx = DIV(scr_xdim, count);
		dy = DIV(scr_ydim, count);
		for (n = 0; (iflag = tty_check()) && n <= dx + dy; n++)
		for (ny = scr_uly; ny <= scr_lry; ny += dy)
		for (nx = scr_ulx; nx <= scr_lrx; nx += dx)
		for (y = 0; y <= n && y < dy; y++) {
			if (n - y < dx)
				scr_pixel(nx + n - y, ny + y);
		}
		break;

	case WW1:			/* windshield washer wipe pivot ul */
		buf = alloca(scr_xdim * sizeof(int));	/* allocate temporary storage */
		for (x = 0; x < scr_xdim; x++)
			buf[x] = 0;
		for (y = 0; (iflag = tty_check()) && y < scr_ydim; y++)
		for (x = 0; x < scr_xdim; x++) {
			if (x * y < (scr_xdim-1) * buf[x])
				continue;
			scr_pixel(scr_ulx + x, scr_uly + buf[x]);
			++buf[x];
		}
		for (n = scr_xdim - 2; iflag && (iflag = tty_check()) && n >= 0; n--)
		for (x = 0; x <= n; x++)
		while (x * (scr_ydim - 1) >= n * buf[x]) {
			scr_pixel(scr_ulx + x, scr_uly + buf[x]);
			if (++buf[x] >= scr_ydim)
				break;
		}
		break;

	case WW2:			/* 4-part propeller wipe */
		count = DIV(scr_xymax, 2) + 1;
		buf = alloca(count * sizeof(int));	/* allocate temporary storage */
		for (x = 0; x < count; x++)
			buf[x] = 0;
		for (y = 0; (iflag = tty_check()) && y < count; y++)
		for (x = 0; x < count; x++) {
			if (x * y < count * buf[x])
				continue;
			scr_pixel(scr_x0 + x, scr_y0 + buf[x]);
			scr_pixel(scr_x0 - buf[x], scr_y0 + x);
			scr_pixel(scr_x0 - x, scr_y0 - buf[x]);
			scr_pixel(scr_x0 + buf[x], scr_y0 - x);
			++buf[x];
		}
		for (n = count - 1; iflag && (iflag = tty_check()) && n >= 0; n--)
		for (x = 0; x <= n; x++)
		while (x * count >= n * buf[x]) {
			scr_pixel(scr_x0 + x, scr_y0 + buf[x]);
			scr_pixel(scr_x0 - buf[x], scr_y0 + x);
			scr_pixel(scr_x0 - x, scr_y0 - buf[x]);
			scr_pixel(scr_x0 + buf[x], scr_y0 - x);
			if (++buf[x] >= count)
				break;
		}
		break;

	case WW3:			/* windshield washer wipe pivot ul/lr */
		buf = alloca(scr_xdim * sizeof(int));	/* allocate temporary storage */
		for (x = 0; x < scr_xdim; x++)
			buf[x] = 0;
		for (y = 0; (iflag = tty_check()) && y < scr_ydim; y++)
		for (x = 0; x < scr_xdim; x++) {
			if (x * y < (scr_xdim-1) * buf[x])
				continue;
			scr_pixel(scr_ulx + x, scr_uly + buf[x]);
			++buf[x];
		}
		for (y = 0; iflag && (iflag = tty_check()) && y < scr_ydim; y++)
		for (x = 0; x < scr_xdim; x++) {
			if ((scr_xdim-1) * (scr_ydim-1 - buf[x]) < (scr_xdim-1 - x) * (scr_ydim-1 - y))
				continue;
			scr_pixel(scr_ulx + x, scr_uly + buf[x]);
			++buf[x];
		}
		break;

	case SPIN:			/* spiral in */
		count = RAND(10, 30);		/* range [10, 30] strips */
		n = DIV(scr_xymin, count);	/* width of a strip */
		ulx = scr_ulx;
		uly = scr_uly;
		lrx = scr_lrx;
		lry = scr_lry;
		for (;;) {
			for (y = uly; (iflag = tty_check()) && y <= lry; y++)
			for (x = ulx; x < ulx + n; x++)
				scr_pixel(x, y);	/* left strip */
			if (!iflag || (ulx += n) > lrx)
				break;
			for (x = ulx; (iflag = tty_check()) && x <= lrx; x++)
			for (y = lry - n + 1; y <= lry; y++)
				scr_pixel(x, y);	/* bottom strip */
			if (!iflag || (lry -= n) < uly)
				break;
			for (y = lry; (iflag = tty_check()) && y >= uly; y--)
			for (x = lrx - n + 1; x <= lrx; x++)
				scr_pixel(x, y);	/* right strip */
			if (!iflag || (lrx -= n) < ulx)
				break;
			for (x = lrx; (iflag = tty_check()) && x >= ulx; x--)
			for (y = uly; y < uly + n; y++)
				scr_pixel(x, y);	/* top strip */
			if (!iflag || (uly += n) > lry)
				break;
		}
		break;

	case SPIN2:			/* double spiral in */
		count = RAND(10, 30);		/* range [10, 30] strips */
		n = DIV(scr_xymin, count);	/* width of a strip */
		ulx = scr_ulx;
		uly = scr_uly;
		lrx = scr_lrx;
		lry = scr_lry;
		for (;;) {
			for (y = 0; (iflag = tty_check()) && y <= lry - uly; y++)
			for (x = 0; x < n; x++) {
				scr_pixel(ulx + x, uly + y);	/* left strip */
				scr_pixel(lrx - x, lry - y);	/* right strip */
			}
			ulx += n;
			lrx -= n;
			if (!iflag || ulx > lrx)
				break;
			for (x = 0; (iflag = tty_check()) && x <= lrx - ulx; x++)
			for (y = 0; y < n; y++) {
				scr_pixel(ulx + x, lry - y);	/* bottom strip */
				scr_pixel(lrx - x, uly + y);	/* top strip */
			}
			lry -= n;
			uly += n;
			if (!iflag || lry < uly)
				break;
		}
		break;

	case CLR:			/* compound effects left, right */
	case CUD:			/* compound effects up, down */
	case CSQ:			/* compound effects 2 x 2 */
		/*
		 * The compound effects pick an effect in the range [1, NWIPES]
		 * and applies it repeatedly in 2 or 4 subwindows.
		 * The range currently excludes circles, compound effects
		 * (applied recursively), and dissolves.
		 * This choice is a purely aesthetic judgement
		 * about what looks good.
		 * The srand(random) kludge makes the output look better when
		 * the applied effect is n-window or nxn-window,
		 * it results in the same count being used in each window.
		 */
		random = rand();		/* a random value */
		i = 1 + random % NWIPES;	/* new effect */
		ulx = scr_ulx;
		uly = scr_uly;
		lrx = scr_lrx;
		lry = scr_lry;		/* save current clipping window */
		if (effect == CLR) {
			scr_clip(ulx, uly, (ulx + lrx) / 2, lry);
			iflag = scr_effect((EFFECT)i);
				scr_clip((ulx + lrx) / 2 + 1, uly, lrx, lry);
			if (iflag) {
				srand(random);
				iflag = scr_effect((EFFECT)i);
			}
		} else if (effect == CUD) {
			scr_clip(ulx, uly, lrx, (uly +lry ) / 2);
			iflag = scr_effect((EFFECT)i);
			scr_clip(ulx, (uly + lry) / 2 + 1, lrx, lry);
			if (iflag) {
				srand(random);
				iflag = scr_effect((EFFECT)i);
			}
		} else {
			scr_clip(ulx, uly, (ulx + lrx) / 2, (uly +lry ) / 2);
			srand(random);
			iflag = scr_effect((EFFECT)i);
			scr_clip((ulx + lrx) / 2 + 1, (uly + lry) / 2 + 1, lrx, lry);
			if (iflag) {
				srand(random);
				iflag = scr_effect((EFFECT)i);
			}
			scr_clip((ulx + lrx) / 2 + 1, uly, lrx, (uly +lry ) / 2);
			if (iflag) {
				srand(random);
				iflag = scr_effect((EFFECT)i);
			}
			scr_clip(ulx, (uly + lry) / 2 + 1, (ulx + lrx) / 2, lry);
			if (iflag) {
				srand(random);
				iflag = scr_effect((EFFECT)i);
			}
		}
		scr_clip(ulx, uly, lrx, lry);	/* restore full window */
		break;

	case CSQSQ:			/* compound effects in nxn windows */
		count = RAND(2, 10);		/* range [2, 10] windows */
		dx = DIV(scr_xdim, count);
		dy = DIV(scr_ydim, count);
		buf = alloca(count * count * sizeof(int));	/* allocate temporary storage */

		/* Randomize values [0, count * count - 1] in buf[]. */
		for (i = 0; i < count * count; i++)
			buf[i] = i;
		for (i = count * count - 1; i > 0; i--) {
			n = rand() % (i + 1);	/* in range [0, i] */
			x = buf[i];
			buf[i] = buf[n];
			buf[n] = x;
		}
		random = rand();		/* a random value */
		i = 1 + random % NWIPES;	/* new effect */
		ulx = scr_ulx;
		uly = scr_uly;
		lrx = scr_lrx;
		lry = scr_lry;			/* save current clipping window */
		for (n = 0; iflag && n < count * count; n++) {
			x = buf[n] % count;
			y = buf[n] / count;
			scr_clip(ulx + x * dx, uly + y * dy,
				ulx + (x + 1) * dx - 1, uly + (y + 1) * dy - 1);
			srand(random);
			iflag = scr_effect((EFFECT)i);
			scr_clip(ulx, uly, lrx, lry);	/* restore full window */
		}
		break;

	case DISS1:				/* dissolves */
	case DISS2:
	case DISS3:
		/*
		 * DISS1 hits every pixel, provided prime is prime
		 * relative to scr_npixels.
		 * DISS2 hits half the pixels and leaves others unchanged,
		 * DISS3 hits either one-third or two-thirds of the pixels.
		 */
		prime = primes[RAND(0, NDISS - 1)];
		i = count = scr_npixels / NCHECK;	/* tty_check() NCHECK times */
		n = (effect == DISS1) ? 0
		  : (effect == DISS2) ? scr_npixels / 2 : scr_npixels / 3;
		do {
			d = div(n, scr_xmax);
			scr_pixel(d.rem, d.quot);
			if ((n += prime) >= scr_npixels)
				n %= scr_npixels;
			if (--i == 0) {
				if (!(iflag = tty_check()))
					break;
				i = count;
			}
		} while (n != 0);
		break;

	case DISSSQ:				/* dissolve in random NSQPIX x NSQPIX squares */
		dx = DIV(scr_xdim, NSQPIX);
		dy = DIV(scr_ydim, NSQPIX);
		nsq = dx * dy;			/* number of squares */

		/* buf[0] ... buf[nsq - 1] is a list of unchanged squares. */
		buf = alloca(nsq * sizeof(int));	/* allocate temporary storage */
		for (i = 0; i < nsq; i++)
			buf[i] = i;

		i = count = nsq / NCHECK;		/* tty_check() NCHECK times */
		do {
			x = rand() % nsq;
			n = buf[x];		/* square to change */
			if (--nsq >= 0)
				buf[x] = buf[nsq];
			d = div(n, dx);
			nx = scr_ulx + NSQPIX * d.rem;
			ny = scr_uly + NSQPIX * d.quot;
			for (y = ny; y < ny + NSQPIX; y++)
			for (x = nx; x < nx + NSQPIX; x++)
				scr_pixel(x, y);
			if (--i == 0) {
				if (!(iflag = tty_check()))
					break;
				i = count;
			}
		} while (nsq > 0);
		break;

	case FADEIN:			/* fade in */
	case FADEOUT:			/* fade out */
		/*
		 * For our purposes here,
		 * a fade out is a dissolve to BLACK, a random color,
		 * or a random color on each pixel,
		 * and a fade in is a fade out followed by a dissolve.
		 * See comments on DISS[123] above.
		 */
		random = rand();
		srand(random);
		flag = random % 3;
		prime = primes[rand() % NDISS];
		if (flag)
			r = (flag == 1) ? BLACK : rand() % scr_ncolors;
		i = count = scr_npixels / NCHECK;	/* tty_check() NCHECK times */
		n = 0;
		do {
			int color;
			char *cp;

			cp = screen + n;
			color = *cp;
			*cp = (flag) ? r : rand() % scr_ncolors;
			d = div(n, scr_xmax);
			scr_pixel(d.rem, d.quot);
			*cp = color;		/* restore current color */
			if ((n += prime) >= scr_npixels)
				n %= scr_npixels;
			if (--i == 0) {
				if (!(iflag = tty_check()))
					break;
				i = count;
			}
		} while (n != 0);
		if (effect == FADEIN) {
			srand(random);
			iflag = scr_effect(DISS1);
		}
		break;

	}
	return iflag;
}

/* end of scr_effect.c */
