/*
 * spiral.c
 * 12/7/93
 * by Stephen A. Ness (steve@mwc.com)
 * Cf. Gombrich, fig. 147: Spyros Horemis, Op art design, 1970.
 * VGA hack, assumes libscreen.a and libvga.a:
 *	cc -VNDP spiral.c -lscreen -lvga -lm
 * The -VNDP not essential but this is slow without it.
 * 640x480x16 VGA screen is wired in for the moment.
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "screen.h"
#include "vga.h"

void	spiral();

/* Defaults. */
#define	COLORS	16		/* number of colors */
#define	NCOLORS	9
#define	SLICES	72
#define	BACKGR	5
#define	OFFSET	0

int	colors, slices, offset, backgr;
double	radsq[1000];			/* radii squared of circles */

main(argc, argv) int argc; char *argv[];
{
	/* Process command line options. */
	colors = (argc > 1) ? atoi(argv[1]) : NCOLORS;
	slices = (argc > 2) ? atoi(argv[2]) : SLICES;
	offset = (argc > 3) ? atoi(argv[3]) : OFFSET;
	backgr = (argc > 4) ? atoi(argv[4]) : BACKGR;
	if (argc > 5
	 || colors <= 0 || colors >  COLORS
	 || slices <= 0 || slices >  1000
	 || offset <  0 || offset >= COLORS
	 || backgr <  0 || backgr >= COLORS) {
		fprintf(stderr, "Sorry...\n");
		fprintf(stderr, "Usage: spiral [ colors [ slices [ offset [ backgr ] ] ] ]\n");
		exit(1);
	}

	/* Display the picture.  The VGA mode is wired in for now. */
	scr_init(G640x480x16);
	spiral();
	scr_display(1);
	while (tty_check()) {
		scr_recolor();
		scr_effect(RANDOM);
	}
	scr_quit();
	exit(0);
}

void
spiral()
{
	register int x, y;
	register char *s1, *s2, *s3, *s4;
	int ysq, rsq, slice, circle, ncircles;
	double yd, theta, R, pislice;

	/* Initialize screen to background color. */
	scr_clear(backgr);

	/*
	 * Compute radii squared of circles and other constants.
	 * The resolution deteriorates at the center,
	 * the inner circle sizes should depend on slices but here do not.
	 */
	R = 1. - (2.*PI) / (slices+PI);
	pislice = slices / (2.*PI);
	for (ncircles = 0, yd = scr_y0 - 1; yd > 10.; yd *= R)
		radsq[ncircles++] = yd * yd;
	for (yd = 8.; yd >= 0.; yd -= 2.)
		radsq[ncircles++] = yd * yd;

	/*
	 * Find which circle and pie slice pixel (x, y) falls in.
	 * Choose a screen color for the pixel accordingly.
	 * The for-loops cover the first quadrant only
	 * because computation of the atan is very slow
	 * (it should probably be cached for efficiency)
	 * and values in the other three quadrants are easily derived.
	 */
	for (y = 0; y < scr_y0; y++) {
		ysq = y * y;
		s1 = s2 = scr0_addr(0,  y);
		s3 = s4 = scr0_addr(0, -y);
		for (x = 0; x < scr_x0; x++) {

			/* Find which circle (x, y) falls in. */
			rsq = x * x + ysq;
			for (circle = 0; radsq[circle] > rsq; circle++)
				;
			if (circle == 0) {			/* outside */
				s1++;				/* use backgr */
				s2--;
				s3++;
				s4--;
				continue;
			} else if (circle >= ncircles - 3) {	/* center */
				*s1++ = *s2-- = *s3++ = *s4-- = (circle + offset) % colors;
				continue;
			}

			/*
			 * theta is in the range [-pi, pi],
			 * so 1+theta/PI is in the range [0, 2]
			 * and (1+theta/PI)*slices/2
			 * == slices/2 + theta * slices / (2 * PI)
			 * is in the range [0, slices];
			 * it cuts the screen into slices pie slices.
			 */
			theta = atan2((double)y, (double)x);
			slice = slices/2 + theta * pislice;
			*s1++ = (slice + circle + offset) % colors;
			slice = slices/2 + (PI - theta) * pislice;
			*s2-- = (slice + circle + offset) % colors;
			slice = slices/2 + (-theta) * pislice;
			*s3++ = (slice + circle + offset) % colors;
			slice = slices/2 + (theta - PI) * pislice;
			*s4-- = (slice + circle + offset) % colors;
		}
	}
}

/* end of spiral.c */
